#!/bin/env python

import yaml
import sys
import logging
import argparse
from os.path import basename
from pprint import pprint
from textwrap import shorten
from dash import Dash
from rocolib.library import getComponent, getComponentTree
from rocolib.api.composables.graph.Joint import FingerJoint


def test(component, params, thickness, outdir=None, display=False, outputs=None):
    f = getComponent(component)
    if params is not None:
        for p in params:
            try:
                v = eval(str(p[1]))
            except NameError:
                # Using a string as a dimensions parameter
                v = str(p[1])
            f.setParameter(p[0], v)
    if thickness is None:
        t = 0
        j = None
    else:
        t = thickness
        j = FingerJoint(thickness=t)

    f.make()

    if outdir:
        filedir = f"{outdir}/{component}"
        outputs = outputs or True
    else:
        filedir = None
        outputs = outputs or ()
    ret = f.makeOutput(outputs = outputs, filedir=filedir, thickness=t, joint=j, remake=False)

    for composable, outs in ret.items():
        if outs:
            print(f"Composable: {composable}")
            pprint({k: shorten(str(v), 60) for k, v in outs.items()})

    if display:
        app = Dash(__name__)
        html = f.visualize(outputs = True, thickness=t, joint=j, remake=False)
        app.layout = html
        app.run_server(debug=True)

def cli_argparse():
    LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    DEFAULT_LOG_LEVEL = "WARNING"

    rocoview = (basename(sys.argv[0]) == "rocoview")

    parser = argparse.ArgumentParser()

    parser.add_argument("component", nargs='?',
                        help="Name of the component you'd like to test")
    parser.add_argument("--file", "-f", type=str, help="Load component and parameters from file")
    parser.add_argument("--output", "-o", type=str, help="Output directory")
    parser.add_argument("--list", "-l", action='store_true',
                        help="List all known library components")

    parser.add_argument("--verbose", "-v", help="Increase logging level (can be used multiple times)",
                        dest="log_level", action="append_const", const=-1)
    parser.add_argument("--quiet", "-q", help="Decrease logging level (can be used multiple times)",
                        dest="log_level", action="append_const", const=1)
    parser.add_argument("-I", help="List component interfaces (use multiple times for more detail)",
                        dest="interface_list", action="append_const", const=1)
    parser.add_argument("-P", help="List component parameters (use multiple times for more detail)",
                        dest="param_list", action="append_const", const=1)
    parser.add_argument("-t", type=float, help="Thickness (i.e. making with wood)")
    if not rocoview:
        parser.add_argument("-d", action='store_true', help="Display visualizations")
    parser.add_argument("-s", type=str, action='append', help="Select specific outputs to save")
    parser.add_argument("-S", help="List composable outputs (use multiple times for more detail)",
                        dest="output_list", action="append_const", const=1)
    parser.add_argument("-p", nargs=2, action='append',
        metavar=("NAME", "VALUE"),
        help="Set component parameter NAME to value VALUE")

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    if rocoview:
        args.d = True

    log_level = LOG_LEVELS.index(DEFAULT_LOG_LEVEL)
    # For each "-q" and "-v" flag, adjust the logging verbosity accordingly
    # making sure to clamp off the value from 0 to 4, inclusive of both
    for adjustment in args.log_level or ():
        log_level = min(len(LOG_LEVELS) - 1, max(log_level + adjustment, 0))
    log_level_name = LOG_LEVELS[log_level]
    logging.basicConfig(level=getattr(logging, log_level_name))

    acted = False
    if args.list:
        t = getComponentTree()
        for i, cs in enumerate(t):
            chunks = [cs[x:x+4] for x in range(0, len(cs), 4)]
            print(f"{i:2}: " + "\n    ".join((" ".join(x) for x in chunks)))
        acted = True

    if args.file:
        with open(args.file) as f:
            config = yaml.safe_load(f)
        args.component = args.component or config.pop("$component", None)
        args.t = args.t or config.pop("$thickness", None)
        if args.p:
            config.update(dict(args.p))
        args.p = list(config.items())

    if args.output:
        if args.output == '-':
            outdir = None
        else:
            outdir = args.output
    else:
        outdir = not rocoview and "output" or None

    if args.component:
        def printList(arg, lsfn, make=False):
            if not arg:
                return
            f = getComponent(args.component)
            if make:
                f.make()
            ls = lsfn(f)
            ind = min(len(arg), len(ls)) - 1
            pprint(ls[ind])
            if not args.p or args.t or args.d:
                exit(0)

        printList(args.interface_list, lambda c : (lambda info : (
            list(info.keys()),
            {k: v["port"] for k, v in info.items()},
            info,
        ))(c.getInterfaceInfo()))

        printList(args.param_list, lambda c : (lambda info : (
            list(info.keys()),
            {k: v["defaultValue"] for k, v in info.items()},
            {k: {x: v[x] for x in ('defaultValue', 'spec')} for k,v in info.items()},
            info,
        ))(c.getParameterInfo()))

        printList(args.output_list, lambda c : (lambda info : (
            {composable: list(outputs.keys()) for composable, outputs in info.items()},
            {composable: {k: {x: v[x] for x in ('stub', 'widget')} for k, v in outputs.items()} for composable, outputs in info.items()},
            info,
        ))(c.listOutputs()), make=True)

        test(args.component, args.p, thickness=args.t, outdir=outdir, display=args.d, outputs=args.s)
        acted = True

    if not acted:
        parser.print_help(sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    cli_argparse()
