from tkinter import *
import math
import numpy
import logging
import plotly.graph_objects as go

from rocolib.api.composables.graph.Drawing import *
from rocolib.api.composables.graph.DrawingEdge import *


log = logging.getLogger(__name__)

### adapted from https://chart-studio.plotly.com/~empet/15276/converting-a-stl-mesh-to-plotly-gomes/#/
def stl2mesh3d(stl_mesh, scale=1000):
    # scale = 1000 back into roco units = mm

    # this function extracts the unique vertices and the lists I, J, K to define a Plotly mesh3d
    p, q, r = stl_mesh.vectors.shape #(p, 3, 3)
    # the array stl_mesh.vectors.reshape(p*q, r) can contain multiple copies of the same vertex;
    # extract unique vertices from all mesh triangles
    vertices, ixr = numpy.unique(stl_mesh.vectors.reshape(p*q, r), return_inverse=True, axis=0)
    i = numpy.take(ixr, [3*k for k in range(p)])
    j = numpy.take(ixr, [3*k+1 for k in range(p)])
    k = numpy.take(ixr, [3*k+2 for k in range(p)])
    x, y, z = scale * vertices.T
    return dict(x=x, y=y, z=z, i=i, j=j, k=k)

def meshfig(stlmesh, scale=1000, color="#ccccff", axes=True):
    colorscale= [[0, color], [1, color]]
    mesh3D = go.Mesh3d(
            **stl2mesh3d(stlmesh, scale),
            flatshading=True,
            colorscale=colorscale,
            name='model',
            showscale=False)
    layout = go.Layout(
                paper_bgcolor='rgba(0,0,0,0)',
                plot_bgcolor='rgba(0,0,0,0)',
                scene_xaxis_visible=axes,
                scene_yaxis_visible=axes,
                scene_zaxis_visible=axes)

    fig = go.Figure(data=[mesh3D], layout=layout)
    fig.update_layout(margin=dict(r=0, l=0, b=0, t=0),
                      scene_aspectmode='data',
                      width=1024)
    fig.data[0].update(lighting=dict(ambient= 0.18,
                                     diffuse= 1,
                                     fresnel=  .1,
                                     specular= 0,
                                     roughness= .1,
                                     facenormalsepsilon=0))
    fig.data[0].update(lightposition=dict(x=3000,
                                          y=3000,
                                          z=10000));

    return fig

def linestyle(e, **kwargs):
    from rocolib.api.composables.graph.DrawingEdge import EdgeType

    width = 3
    angle = e["angle"]

    if e["interior"]: # Interior decorations
        width = 1
        if e["edgeType"] in ( EdgeType.BOUNDARY_CUT, EdgeType.INTERIOR_CUT ):
            color = 'red'
            zorder = 30
        elif e["edgeType"] in ( EdgeType.FOLD, EdgeType.FLEX ):
            color = 'blue'
            zorder = 20
        elif e["edgeType"] in ( EdgeType.FLAT ):
            color = 'white'
            zorder = 10
        else: # Other?
            color = 'gray'
            zorder = 100
    elif e["faces"] == 1: # CUT
        color = 'red'
        zorder = 30
    elif e["faces"] == 2: # FOLD or FLEX or BEND
        if angle: # FOLD
            if angle > 0: # mountain fold
                color = 'blue'
                zorder = 20
            else: # valley fold
                color = 'green'
                zorder = 20
        else: # FLEX or BEND or FLAT
            color = 'white'
            zorder = 10
    else: # 0 or 3+ faces?  dunno.
        color = 'grey'
        zorder = 100

    ret = dict(color = color, width = width, zorder = zorder)
    ret.update(kwargs)
    return ret

def cutfig(edgeset, scale=1000, axes=True):
    layout = go.Layout(
        showlegend=False,
        dragmode='pan',
        yaxis=dict(scaleanchor="x", scaleratio=1),
        margin=go.layout.Margin(
            l=0, #left margin
            r=0, #right margin
            b=0, #bottom margin
            t=0  #top margin
        )
    )

    fig = go.Figure(layout=layout)
    traces = []

    for n, e in edgeset.items():
        # Adding a trace with a fill, setting opacity to 0
        x0, y0 = e["p0"]
        x1, y1 = e["p1"]
        dx, dy = x1 - x0, y1 - y0
        length = (dx * dx + dy * dy) ** 0.5
        ddx = -dy/20 # hit box width = 10% length
        ddy = dx/20 # hit box width = 10% length

        xs = [x0 - ddx, x0 + ddx, x1 + ddx, x1 - ddx]
        ys = [y0 - ddy, y0 + ddy, y1 + ddy, y1 - ddy]

        hoverdata = "<br>".join([
            f"<b>Length</b>: {length}",
            f"<b>Angle</b>: {e['angle']}",
            f"<b>Faces</b>: {e['faces']}",
            f"<b>Type</b>: {e['edgeType']}",
        ])

        style = linestyle(e)
        zorder = style.pop("zorder")

        traces.append((zorder,
            go.Scatter(
                x = [x0, x1],
                y = [y0, y1],
                mode='lines',
                name=n,
                line=style,
            )
        ))

        style["width"] = 0
        traces.append((zorder,
            go.Scatter(
                x = xs,
                y = ys,
                fill="toself",
                mode='lines',
                name=n,
                text=hoverdata,
                opacity=0,
                line=style,
            )
        ))

    for z, t in sorted(traces, key = lambda x : x[0]):
        fig.add_trace(t)

    return fig
