import numpy as np

from rocolib.utils.roco2sim.material_data import Material as mat, MaterialConst as mc
from rocolib.utils.roco2sim.material_data import materialData as md
from rocolib.utils.roco2sim.format_3d import format_wrl as wrl
from rocolib.utils.roco2sim.format_3d import wrl2stl

"""BaseNodes (No children allowed)"""

class AppearanceNode:
    """appearance"""

    def __init__(self, material: mat = md[mc.MAT_DEFAULT]):
        # if no material specified, use default material
        self._material = material

    def getMaterial(self):
        return self._material

    def getRGB(self):
        return self._material.rgb

    def getRoughness(self):
        return self._material.roughness

    def getReflection(self):
        return self._material.reflection

    def getName(self):
        return self._material.name

    def getDescription(self):
        return self._material.description


class PhysicsNode:
    def __init__(self, material: mat = md[mc.MAT_DEFAULT], mass=0.3, contact=None):
        self.material = material
        self.mass = mass
        self.contact = contact

    def isMassSpecified(self):
        return self.mass is not None



class DeviceNode:
    def __int__(self, name=None, type=None, options=None):
        if options is None:
            options = {}
        self.options = options
        self.name = name
        self.type = type


"""Composable Nodes (can have children)"""
class BaseComposableNode:
    def __init__(self, name, children=None):
        self.name = name
        if children is None:
            children = []
        self.children = children


class ShapeNode(BaseComposableNode):
    """Shape node holds faces and appearance information.
    If solidified, then includes physics"""

    def __init__(self, name, wrl: wrl, trans, appearance: AppearanceNode = None, solidify: PhysicsNode = None,
                 children=None):
        super().__init__(name, children)
        self.trans = trans
        self.wrl = wrl
        self.solid = False


        # if no appearance specified, default
        if appearance is None:
            self.appearance = AppearanceNode()
        else:
            self.appearance = appearance

        # physics is None, then inherit, if Physics node is given, then use it.
        # otherwise not physics
        self.physics = PhysicsNode
        if solidify is None:
            self.solidify()
        elif isinstance(solidify, PhysicsNode):
            self.solidifyWithNewPhysics(solidify)

    def solidify(self):
        """Add physics property to this shape based on appearance material"""
        self.physics = PhysicsNode(self.appearance.getMaterial())
        self.solid = True

    def solidifyWithNewPhysics(self, physics: PhysicsNode):
        """Add physics property to this shape by specified physics nodes"""
        self.physics = physics
        self.solid = True

    def gasify(self):
        self.solid = False

    def isSolid(self):
        return self.solid

    def getXyz(self):
        return self.trans
    def getRotationVector(self):
        R = self.trans[0:3,0:3]
        if np.array_equal(R, np.identity(3)):
            rVector = np.array([0,0,0,0])
        else:
            v = np.array([R[2,1]-R[1,2],R[0,2]-R[2,0],R[1,0]-R[0,1]])
            rVector = np.append(v/np.linalg.norm(v), np.arccos((np.trace(R)-1)/2))

        return rVector


    def getPhysicsNode(self):
        return self.physics

    def getWRL(self):
        return self.wrl

    def getSTL(self):
        return wrl2stl(self.wrl)

    # def getFaces(self):
    #     return self.faces
    # def getAppearance(self):
    #     return self.appearance


class ModuleNode(BaseComposableNode):
    def __init__(self, name, shape: ShapeNode, device: DeviceNode = None, children=None):
        super().__init__(name, children)
        self.shape = shape
        self.device = device

    def getDeviceOptions(self, optionName):
        if optionName in self.device.options.items():
            return self.device.options[optionName]
        return ""

    def getPhysicsNode(self):
        return self.shape.physics
